/*
 * Scott Schneider
 *
 * complex.cpp
 */

#include <iostream>
using namespace std;

#include "work_unit.h"
#include "factory.h"

const int N = 1024;

factory<mixed_work> matrixops(LOGICAL, LOCAL);

class strassen_base: public task_unit
{
	protected:
		double* a;
		double* b;
		double* c;
		int matrix_size;
		int a_width;
		int b_width;
		int c_width;

	public:
		void work_init(double* _a, double* _b, double* _c, int size, int aw, int bw, int cw, task_unit* parent)
		{
			task_unit::work_init(parent);
			a = _a;
			b = _b;
			c = _c;
			matrix_size = size;
			a_width = aw;
			b_width = bw;
			c_width = cw;
		}
};

class matrix_add: public strassen_base 
{	
	public:
		void work()
		{
			for (int i = 0; i < matrix_size; ++i) {
				for (int j = 0; j < matrix_size; ++j) {
					c[i * c_width + j] = a[i * a_width + j] + b[i * b_width + j];
				}
			}
		}
};

class matrix_subtract: public strassen_base 
{
	public:
		void work()
		{
			for (int i = 0; i < matrix_size; ++i) {
				for (int j = 0; j < matrix_size; ++j) {
					c[i * c_width + j] = a[i * a_width + j] - b[i * b_width + j];
				}
			}
		}
};

class strassen: public strassen_base
{
	public:
		void work()
		{
			// base case
			if (matrix_size <= 64) {
				for (int i = 0; i < matrix_size; ++i) {
					for (int j = 0; j < matrix_size; ++j) {
						double sum = 0;
						for (int k = 0; k < matrix_size; ++k) {
							sum += a[i * a_width + k] * b[k * b_width + j];
						}
						c[i * c_width + j] = sum;
					}
				}
			}
			else {
				const int quad_size = matrix_size >> 1;

				double* a11 = a;
				double* a12 = a + quad_size;
				double* a21 = a + (a_width * quad_size);
				double* a22 = a + (a_width * quad_size) + quad_size;
				
				double* b11 = b;
				double* b12 = b + quad_size;
				double* b21 = b + (b_width * quad_size);
				double* b22 = b + (b_width * quad_size) + quad_size;

				double* c11 = c;
				double* c12 = c + quad_size;
				double* c21 = c + (c_width * quad_size);
				double* c22 = c + (c_width * quad_size) + quad_size;

				double* s1 = new double[quad_size * quad_size];
				double* s2 = new double[quad_size * quad_size];
				double* s3 = new double[quad_size * quad_size];
				double* s4 = new double[quad_size * quad_size];

				double* t1 = new double[quad_size * quad_size];
				double* t2 = new double[quad_size * quad_size];
				double* t3 = new double[quad_size * quad_size];
				double* t4 = new double[quad_size * quad_size];

				double* p1 = new double[quad_size * quad_size];
				double* p2 = new double[quad_size * quad_size];
				double* p3 = new double[quad_size * quad_size];
				double* p4 = new double[quad_size * quad_size];
				double* p5 = new double[quad_size * quad_size];
				double* p6 = new double[quad_size * quad_size];
				double* p7 = new double[quad_size * quad_size];

				double* u2 = new double[quad_size * quad_size];
				double* u3 = new double[quad_size * quad_size];
				double* u6 = new double[quad_size * quad_size];

				// additions and subtractions
				barrier_object work_s1 = matrixops.spawn<matrix_add>(a21, a22, s1, quad_size, a_width, a_width, quad_size, this);
				barrier_object work_t1 = matrixops.spawn<matrix_subtract>(b12, b11, t1, quad_size, b_width, b_width, quad_size, this);
				matrixops.spawn<matrix_subtract>(a11, a21, s3, quad_size, a_width, a_width, quad_size, this);	// work_s3
				matrixops.spawn<matrix_subtract>(b22, b12, t3, quad_size, b_width, b_width, quad_size, this);	// work_t3

				matrixops.parent_barrier(work_s1, q);
				barrier_object work_s2 = matrixops.spawn<matrix_subtract>(s1, a11, s2, quad_size, quad_size, a_width, quad_size, this);

				matrixops.parent_barrier(work_t1, q);
				barrier_object work_t2 = matrixops.spawn<matrix_subtract>(b22, t1, t2, quad_size, b_width, quad_size, quad_size, this);

				matrixops.parent_barrier(work_s2, q);
				matrixops.spawn<matrix_subtract>(a12, s2, s4, quad_size, a_width, quad_size, quad_size, this); // work_s4

				matrixops.parent_barrier(work_t2, q);
				matrixops.spawn<matrix_subtract>(b21, t2, t4, quad_size, b_width, quad_size, quad_size, this); // work_t4

				matrixops.child_barrier(this);

				// multiplications
				matrixops.spawn<strassen>(a11, b11, p1, quad_size, a_width, b_width, quad_size, this);		// work_p1
				matrixops.spawn<strassen>(a12, b21, p2, quad_size, a_width, b_width, quad_size, this);		// work_p2
				matrixops.spawn<strassen>(s1, t1, p3, quad_size, quad_size, quad_size, quad_size, this);	// work_p3
				matrixops.spawn<strassen>(s2, t2, p4, quad_size, quad_size, quad_size, quad_size, this);	// work_p4
				matrixops.spawn<strassen>(s3, t3, p5, quad_size, quad_size, quad_size, quad_size, this);	// work_p5
				matrixops.spawn<strassen>(s4, b22, p6, quad_size, quad_size, b_width, quad_size, this);		// work_p6
				matrixops.spawn<strassen>(a22, t4, p7, quad_size, a_width, quad_size, quad_size, this);		// work_p7

				matrixops.child_barrier(this);

				// final additions and assignment
				matrixops.spawn<matrix_add>(p1, p2, c11, quad_size, quad_size, quad_size, c_width, this); 	// work_u1
				barrier_object work_u2 = matrixops.spawn<matrix_add>(p1, p4, u2, quad_size, quad_size, quad_size, quad_size, this);

				matrixops.parent_barrier(work_u2, q);
				barrier_object work_u3 = matrixops.spawn<matrix_add>(u2, p5, u3, quad_size, quad_size, quad_size, quad_size, this);
				barrier_object work_u6 = matrixops.spawn<matrix_add>(u2, p3, u6, quad_size, quad_size, quad_size, quad_size, this);

				matrixops.parent_barrier(work_u3, q);
				matrixops.spawn<matrix_add>(u3, p3, c22, quad_size, quad_size, quad_size, c_width, this);	// work_u5
				matrixops.spawn<matrix_add>(u3, p7, c21, quad_size, quad_size, quad_size, c_width, this);	// work_u4

				matrixops.parent_barrier(work_u6, q);
				matrixops.spawn<matrix_add>(u6, p6, c12, quad_size, quad_size, quad_size, c_width, this);	// work_u7

				matrixops.work_end_barrier(this);

				delete s1; delete s2; delete s3; delete s4;

				delete t1; delete t2; delete t3; delete t4;

				delete p1; delete p2; delete p3; delete p4;
				delete p5; delete p6; delete p7; 

				delete u2; delete u3; delete u6;
			}
		}
};

int main(int argc, char** argv)
{
	double* a = new double[N * N];
	double* b = new double[N * N];
	double* c = new double[N * N];

	for (int i = 0; i < N; ++i) {
		for (int j = 0; j < N; ++j) {
			a[i * N + j] = b[i * N + j] = 3;
			c[i * N + j] = 0;
		}
	}

	matrixops.start_working();
	barrier_object root = matrixops.spawn<strassen>(a, b, c, N, N, N, N, (strassen*)NULL);
	matrixops.root_barrier(root);
	matrixops.stop_working();

	if (argc == 2) {
		double* check = new double[N * N];

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				double sum = 0;
				for (int k = 0; k < N; ++k) {
					sum += a[i * N + k] * b[k * N + j];
				}
				check[i * N + j] = sum;
			}
		}

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				if ((c[i * N + j] >= (check[i * N + j]+0.05)) || (c[i * N + j] <= (check[i * N + j]-0.05))) {
					cout	<< "    c[" << i << "][" << j << "] = " << c[i * N + j] << endl
						<< "check[" << i << "][" << j << "] = " << check[i * N + j] << endl
						<< "Incorrect." << endl;
					exit(1);
				}
			}
		}
		cout << "Correct." << endl;
	}

	return 0;
}
