/*
 * Scott Schneider
 *
 * complex.cpp
 */

#include <iostream>
using namespace std;

#include <omp.h>

const int N = 2048;

void strassen(double* a, double* b, double* c, int matrix_size, int a_width, int b_width, int c_width)
{
	// base case
	if (matrix_size <= 64) {
		for (int i = 0; i < matrix_size; ++i) {
			for (int j = 0; j < matrix_size; ++j) {
				double sum = 0;
				for (int k = 0; k < matrix_size; ++k) {
					sum += a[i * a_width + k] * b[k * b_width + j];
				}
				c[i * c_width + j] = sum;
			}
		}
	}
	else {
		const int quad_size = matrix_size >> 1;
		int i;
		int j;

		double* a11 = a;
		double* a12 = a + quad_size;
		double* a21 = a + (a_width * quad_size);
		double* a22 = a + (a_width * quad_size) + quad_size;
		
		double* b11 = b;
		double* b12 = b + quad_size;
		double* b21 = b + (b_width * quad_size);
		double* b22 = b + (b_width * quad_size) + quad_size;

		double* c11 = c;
		double* c12 = c + quad_size;
		double* c21 = c + (c_width * quad_size);
		double* c22 = c + (c_width * quad_size) + quad_size;

		double* start_heap = new double[quad_size * quad_size * 18];
		double* heap = start_heap;

		double* s1 = heap; heap += quad_size * quad_size;
		double* s2 = heap; heap += quad_size * quad_size;
		double* s3 = heap; heap += quad_size * quad_size;
		double* s4 = heap; heap += quad_size * quad_size;

		double* t1 = heap; heap += quad_size * quad_size;
		double* t2 = heap; heap += quad_size * quad_size;
		double* t3 = heap; heap += quad_size * quad_size;
		double* t4 = heap; heap += quad_size * quad_size;

		double* p1 = heap; heap += quad_size * quad_size;
		double* p2 = heap; heap += quad_size * quad_size;
		double* p3 = heap; heap += quad_size * quad_size;
		double* p4 = heap; heap += quad_size * quad_size;
		double* p5 = heap; heap += quad_size * quad_size;
		double* p6 = heap; heap += quad_size * quad_size;
		double* p7 = heap; heap += quad_size * quad_size;

		double* u2 = heap; heap += quad_size * quad_size;
		double* u3 = heap; heap += quad_size * quad_size;
		double* u6 = heap; heap += quad_size * quad_size;

		#pragma omp parallel for private(i, j)
		for (i = 0; i < quad_size; ++i) {
			for (j = 0; j < quad_size; ++j) {
				s1[i * quad_size + j] = a21[i * a_width + j] + a22[i * a_width + j];
				t1[i * quad_size + j] = b12[i * b_width + j] - b11[i * b_width + j];
				s3[i * quad_size + j] = a11[i * a_width + j] - a21[i * a_width + j];
				t3[i * quad_size + j] = b22[i * b_width + j] - b12[i * b_width + j];

				s2[i * quad_size + j] = s1[i * quad_size + j] - a11[i * a_width + j];
				t2[i * quad_size + j] = b22[i * b_width + j] - t1[i * quad_size + j];
				s4[i * quad_size + j] = a12[i * a_width + j] - s2[i * quad_size + j];
				t4[i * quad_size + j] = b21[i * b_width + j] - t2[i * quad_size + j];
			}
		}

		#pragma omp parallel sections 
		{
			#pragma omp section 
			strassen(a11, b11, p1, quad_size, a_width, b_width, quad_size);

			#pragma omp section 
			strassen(a12, b21, p2, quad_size, a_width, b_width, quad_size);

			#pragma omp section 
			strassen(s1, t1, p3, quad_size, quad_size, quad_size, quad_size);
			
			#pragma omp section 
			strassen(s2, t2, p4, quad_size, quad_size, quad_size, quad_size);

			#pragma omp section 
			strassen(s3, t3, p5, quad_size, quad_size, quad_size, quad_size);

			#pragma omp section 
			strassen(s4, b22, p6, quad_size, quad_size, b_width, quad_size);

			#pragma omp section 
			strassen(a22, t4, p7, quad_size, a_width, quad_size, quad_size);
		}

		#pragma omp parallel for private(i, j)
		for (i = 0; i < quad_size; ++i) {
			for (j = 0; j < quad_size; ++j) {
				u2[i * quad_size + j] = p1[i * quad_size + j] + p4[i * quad_size + j];
				u3[i * quad_size + j] = u2[i * quad_size + j] + p5[i * quad_size + j];
				u6[i * quad_size + j] = u2[i * quad_size + j] + p3[i * quad_size + j];

				c11[i * c_width + j] = p1[i * quad_size + j] + p2[i * quad_size + j];
				c12[i * c_width + j] = u6[i * quad_size + j] + p6[i * quad_size + j];
				c21[i * c_width + j] = u3[i * quad_size + j] + p7[i * quad_size + j];
				c22[i * c_width + j] = u3[i * quad_size + j] + p3[i * quad_size + j];
			}
		}
		
		delete start_heap;
	}
}

int main(int argc, char** argv)
{
	double* a = new double[N * N];
	double* b = new double[N * N];
	double* c = new double[N * N];

	if (argc < 2) {
		cout << "omp_complex <num_threads> [check]" << endl;
		exit(2);
	}

	omp_set_num_threads(atoi(argv[1]));

	for (int i = 0; i < N; ++i) {
		for (int j = 0; j < N; ++j) {
			a[i * N + j] = b[i * N + j] = 3;
			c[i * N + j] = 0;
		}
	}

	strassen(a, b, c, N, N, N, N);

	if (argc == 3) {
		double* check = new double[N * N];

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				double sum = 0;
				for (int k = 0; k < N; ++k) {
					sum += a[i * N + k] * b[k * N + j];
				}
				check[i * N + j] = sum;
			}
		}

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				if ((c[i * N + j] >= (check[i * N + j]+0.05)) || (c[i * N + j] <= (check[i * N + j]-0.05))) {
					cout	<< "    c[" << i << "][" << j << "] = " << c[i * N + j] << endl
						<< "check[" << i << "][" << j << "] = " << check[i * N + j] << endl
						<< "Incorrect." << endl;
					exit(1);
				}
			}
		}
		cout << "Correct." << endl;
	}

	return 0;
}
