/*
 * rec_parallel.cpp
 *
 * Scott Schneider
 *
 * The code for the recursive matrix multiplication was taken from matmul.cilk, 
 * which is part of the Cilk 5.3.2 package:
 * 	http://supertech.lcs.mit.edu/cilk/
 */

#include "factory.h"
#include "work_unit.h"

#include <iostream>
using namespace std;

const int N = 2000; 

class rec_matmul: public task_unit {

	private:
		int m;
		int n;
		int p;
		int ld;
		bool add;
		double *one;
		double *two;
		double *result;
		factory<rec_matmul>* fac;

	public:
		rec_matmul():
			m(0),
			n(0),
			p(0),
			ld(0),
			add(false),
			one(NULL),
			two(NULL),
			result(NULL),
			fac(NULL)
			{}

		void work_init(	int _m, int _n, int _p, int _ld, bool _add, 
				double *_one, double *_two, double *_result,
				factory<rec_matmul>* _fac, 
				rec_matmul* parent)
		{
			task_unit::work_init(parent);
			m = _m;
			n = _n;
			p = _p;
			ld = _ld;
			add = _add;
			one = _one;
			two = _two;
			result = _result;
			fac = _fac;
		}

		void work()
		{
			if ((m + n + p) <= 64) {
				// base case
				if (add) {
					for (int i = 0; i < m; ++i)
						for (int k = 0; k < p; ++k) {
							double sum = 0.0;
							for (int j = 0; j < n; ++j)
								sum += one[i * ld + j] * two[j * ld + k];
							result[i * ld + k] += sum;
						}
				} else {
					for (int i = 0; i < m; ++i)
						for (int k = 0; k < p; ++k) {
							double sum = 0.0;
							for (int j = 0; j < n; ++j)
								sum += one[i * ld + j] * two[j * ld + k];
							result[i * ld + k] = sum;
		    				}
	  			}
			}
			else if (m >= n && n >= p) {
				int m1 = m >> 1;
				fac->spawn(m1, n, p, ld, add, one, two, result, fac, this);
				fac->spawn(m - m1, n, p, ld, add, (double*)(one + m1 * ld), two, (double*)(result + m1 * ld), fac, this);
			}
			else if (n >= m && n >= p) {
				int n1 = n >> 1;
				fac->spawn(m, n1, p, ld, add, one, two, result, fac, this);
				fac->child_barrier(this);
				fac->spawn(m, n - n1, p, ld, true, (double*)(one + n1), (double*)(two + n1 * ld), result, fac, this);
			}
			else {
				int p1 = p >> 1;
				fac->spawn(m, n, p1, ld, add, one, two, result, fac, this);
				fac->spawn(m, n, p - p1, ld, add, one, (double*)(two + p1), (double*)(result + p1), fac, this);
			}
			fac->work_end_barrier(this);
		}
};

int main(int argc, char** argv)
{
	factory<rec_matmul>* multiply_factory	= new factory<rec_matmul>(LOGICAL, STEAL);
	double (*one)[N]			= new double[N][N];
	double (*two)[N]			= new double[N][N];
	double (*result)[N]			= new double[N][N];

	for (int i = 0; i < N; ++i) {
		for (int j = 0; j < N; ++j) {
			one[i][j] = two[i][j] = 3;
			result[i][j] = 0;
		}
	}

	multiply_factory->register_class<rec_matmul>();

	multiply_factory->start_working();
	barrier_object root = multiply_factory->spawn(N, N, N, N, false, &one[0][0], &two[0][0], &result[0][0], multiply_factory, (rec_matmul*)NULL);
	multiply_factory->root_barrier(root);
	multiply_factory->stop_working();

	// check correctness
	if (argc == 2) {
		double (*check)[N] = new double[N][N];

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				double sum = 0;
				for (int k = 0; k < N; ++k) {
					sum += one[i][k] * two[k][j];
				}
				check[i][j] = sum;
			}
		}

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				if ((result[i][j] >= (check[i][j]+0.05)) || (result[i][j] <= (check[i][j]-0.05))) {
					cout	<< "result[" << i << "][" << j << "] = " << result[i][j] << endl
						<< " check[" << i << "][" << j << "] = " << check[i][j] << endl
						<< "Incorrect." << endl;
					delete multiply_factory;
					exit(1);
				}
			}
		}
		cout << "Correct." << endl;
	}

	delete multiply_factory;
	return 0;
}
