#ifndef __C_QUEUES_H_
#define __C_QUEUES_H_

#define	QUEUE_LEN	(100*1024)

#include "locks.h"
#include "atomic.h"

typedef exp_spin_lock QLock;

struct lf_queue_elem_t {
	void*				_dummy_second;
	volatile struct lf_queue_elem_t	*next;
	void 				*elem;
};

struct lb_fifo_queue_t {
	volatile void 		*queue[QUEUE_LEN];
	volatile unsigned long 	head;
	volatile unsigned long 	tail;
	QLock			lock;
};

struct lf_fifo_queue_t {
	volatile struct lf_queue_elem_t	*head;
	volatile unsigned long		ocount;
	volatile struct lf_queue_elem_t	*tail;
	volatile unsigned long		icount;	
	struct lf_queue_elem_t		_dummy;
};

struct lb_lifo_queue_t {
	volatile void 		*queue[QUEUE_LEN];
	volatile unsigned long 	top;
	QLock			lock;
};

struct lf_lifo_queue_t {
	volatile struct lf_queue_elem_t	*top;
	volatile unsigned long		ocount;
};

template <class Q>
void queue_end(Q* q)
{
	free(q);
}

inline void lf_fifo_queue_init(lf_fifo_queue_t **queue)
{
	*queue = (lf_fifo_queue_t *)malloc(sizeof (lf_fifo_queue_t));
	(*queue)->head = &((*queue)->_dummy);
	(*queue)->tail = &((*queue)->_dummy);
	(*queue)->ocount = 0;
	(*queue)->icount = 0;
	(*queue)->_dummy.next = NULL;
}

inline int lf_fifo_enqueue(lf_fifo_queue_t *queue, void *element)
{
	struct lf_queue_elem_t 	*tail;
	unsigned long		ic;

	((struct lf_queue_elem_t *)element)->next = NULL;
	while(1) {
		ic = queue->icount;
		tail = (struct lf_queue_elem_t*)queue->tail;
		if (compare_and_swap(&(tail->next), NULL, element))
			break;
		else
			compare_and_swap8b(&(queue->tail), tail, ic, tail->next, ic+1);
	}
	compare_and_swap8b(&(queue->tail), tail, ic, element, ic+1);
	return(0);
}

inline void* lf_fifo_dequeue(lf_fifo_queue_t *queue)
{
	unsigned long ic, oc;
	struct lf_queue_elem_t	*head, *next, *retval;
	
	while(1) {
		oc = queue->ocount;
		ic = queue->icount;
		retval = head = (struct lf_queue_elem_t*)queue->head;
		next = (struct lf_queue_elem_t*)head->next;
		if (oc == queue->ocount) {
			if (head == queue->tail) {
				if (next == NULL) {
					return(NULL);
				}
				compare_and_swap8b(&(queue->tail), head, ic, next, ic+1);
			}
			else if (next != NULL) {
				retval = next;
				if (compare_and_swap8b(&(queue->head), head, oc, next, oc+1))
					break;
			}
		}
	}
	return(retval);
}

inline void lb_fifo_queue_init(lb_fifo_queue_t **queue)
{
	*queue = (lb_fifo_queue_t *)malloc(sizeof(lb_fifo_queue_t));
	(*queue)->head = (*queue)->tail = 0;
}

inline int lb_fifo_enqueue(lb_fifo_queue_t *queue, void *element)
{
	queue->lock.lock();
	if (queue->tail == (queue->head+1) % QUEUE_LEN) {
		queue->lock.unlock();
		return(1);
	}
	queue->queue[queue->head] = element;
	queue->head = (queue->head + 1) % QUEUE_LEN;
	queue->lock.unlock();
	return(0);
}

inline void* lb_fifo_dequeue(lb_fifo_queue_t *queue)
{
	void *element;
	
	queue->lock.lock();
	if (queue->head == queue->tail) {
		queue->lock.unlock();
		return(NULL);
	}
	element = (void *)queue->queue[queue->tail];
	queue->tail = (queue->tail + 1) % QUEUE_LEN;
	queue->lock.unlock();
	return(element);
}

inline void lb_lifo_queue_init(lb_lifo_queue_t **queue)
{
	*queue = (lb_lifo_queue_t *)malloc(sizeof(lb_lifo_queue_t));
	(*queue)->top = 0;
}

inline int lb_lifo_enqueue(lb_lifo_queue_t *queue, void *element)
{
	queue->lock.lock();
	if (queue->top == QUEUE_LEN) {
		queue->lock.unlock();
		return(1);
	}
	queue->queue[queue->top] = element;
	queue->top = queue->top + 1;
	queue->lock.unlock();
	return(0);
}

inline void* lb_lifo_dequeue(lb_lifo_queue_t *queue)
{
	void *element;
	
	queue->lock.lock();
	if (queue->top == 0) {
		queue->lock.unlock();
		return(NULL);
	}
	queue->top = queue->top - 1;
	element = (void *)queue->queue[queue->top];
	queue->lock.unlock();
	return(element);
}

inline void lf_lifo_queue_init(lf_lifo_queue_t **queue)
{
	*queue = (lf_lifo_queue_t *)malloc(sizeof (lf_lifo_queue_t));
	(*queue)->top = NULL;
	(*queue)->ocount = 0;
}

inline int lf_lifo_enqueue(lf_lifo_queue_t *queue, void *element)
{
	struct lf_queue_elem_t 	*top;
	unsigned long 		ocount;
	
	while(1) {
		ocount = queue->ocount;
		top = (struct lf_queue_elem_t*)queue->top;
		((struct lf_queue_elem_t *)element)->next = (volatile struct lf_queue_elem_t*)top;
		if (compare_and_swap8b(&(queue->top), top, ocount, element, ocount+1)) {
			return(0);
		}
	}
}

inline void* lf_lifo_dequeue(lf_lifo_queue_t *queue)
{
	struct lf_queue_elem_t 	*head, *next;
	unsigned long 		oc;

	while(1) {
		head = (struct lf_queue_elem_t*)queue->top;
		oc = queue->ocount;
		if (head == NULL)
			return(NULL);
		next = (struct lf_queue_elem_t*)head->next;
		if (compare_and_swap8b(&(queue->top), head, oc, next, oc+1)) {
			return((void *)head);
		}
	}
}

#endif	// __C_QUEUES_H_

