/*
 * factory.h
 *
 * Scott Schneider
 */

#ifndef __FACTORY_H__
#define __FACTORY_H__

#include <queue>
#include <stack>
#include <cassert>
#include "work_unit.h"
#include "slab_allocator.h"
//#include "lf_allocator.h"
//#include "ccmm_allocator.h"
#include "posix_thread.h"
#include "simple_deque.h"
#include "simple_list.h"
#include "queue_wrappers.h"
#include "locks.h"
#include "timing.h"
#include "bindme.h"

/*
 * TODO:
 *
 * - Determine if done[q]/work[q].empty() is more appropriate.
 *
 * - Find a way of keeping the added/executed stuff in the code without
 *   that doesn't use comments, #defines, or add run-time cost.
 */

const int LOGICAL	= 1;
const int PHYSICAL	= 2;

const int PARENT	= 0; 
const int CHILD		= 1; 

typedef noop_lock	FLock;
typedef slab_object	barrier_object;

enum loop_policy	{ STATIC };

enum policy	{ LOCAL, STEAL, STEAL_SMT };

/*
 * A factory is where workers (Threads) do work.
 */
template<class Work>
class factory
{

	struct factory_adapter
	{
		factory<Work>*	ptr;

		factory_adapter(factory<Work>* p): ptr(p) {}
		void operator()(const unsigned int _1, const unsigned int _2)
		{
			ptr->work_cache.local_init();
			ptr->schedule(_1, _2);
		}
	};
		
	struct schedule_args
	{
		factory_adapter			function;
		unsigned int			_1;		// cpu
		unsigned int			_2;		// q
	};

	typedef posix_thread<schedule_args>		Thread;
	//typedef std_lifo_queue<slab_object>		WorkQueue;
	typedef lf_lifo_queue				WorkQueue;
	typedef simple_list<slab_object>		WorkList;
	typedef simple_list<Thread*>			ThreadList;

	private:
		WorkQueue*			work;
		FLock*				work_access;
		ThreadList			threads;
		general_manager			work_cache;
		unsigned int			num_cpus;
		unsigned int			num_threads;
		unsigned int			q_type;
		unsigned long			q_assign;
		volatile unsigned int		ref_count;
		volatile bool			sense;
		volatile bool*			local_sense;
		bool*				done;
		volatile bool*			gone_to_sleep;
		volatile bool			not_sleeped;
		volatile bool			not_joined;

		//unsigned long			added;
		//unsigned long			executed;
		
		slab_object (*policy_alg)(WorkQueue*, FLock*, unsigned int, unsigned int);

		void construct(const unsigned int request_q, const unsigned request_threads, const policy);
		void internal_spawn(slab_object task, int q);
		void barrier(slab_object, const int condition, const int q);
		void long_barrier();
		void long_sleep();

	public:
		factory(const unsigned int request_q, const policy);
		factory(const unsigned int request_q, const unsigned int request_threads, const policy);
		~factory();

		template <class T>
		void register_class()
		{
			work_cache.template register_class<T>();
		}

		void start_working();
		void sleep();
		void wakeup();
		void stop_working();
		void change_policy(const policy p);

		int get_num_threads();

		void root_barrier(slab_object);
		void parent_barrier(slab_object, int);
		void work_end_barrier(tree_unit*);
		void child_barrier(tree_unit*);
		void child_barrier(slab_object, int);
		void barrier();
		void simple_barrier(work_unit*);
		void schedule(const unsigned int cpu, const unsigned int q);

		slab_object spawn(Work* w)
		{
			int q = 0;
			if (w && w->get_q() != -1) {
				q = w->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}
			slab_object task = w;

			internal_spawn(task, q);

			return task;
		}

		template <class T>
		slab_object spawn()
		{
			int q = q_assign++ % num_threads;
			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());
			
			wrk->work_init();
			internal_spawn(task, q);

			return task;
		}
		
		template <class T>
		slab_object spawn(work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1>
		slab_object spawn(_1 arg1, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());
			
			wrk->work_init(arg1, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2>
		slab_object spawn(_1 arg1, _2 arg2, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2, class _3>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2, class _3, class _4>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}
	
		template <class T, class _1, class _2, class _3, class _4, class _5>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}
		
		template <class T, class _1, class _2, class _3, class _4, class _5, class _6>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());
			
			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, static_cast<T*>(parent));

			internal_spawn(task, q);

			return task;
		}
	
		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, static_cast<T*>(parent));
			internal_spawn(task, q);
			
			return task;
		}
		
		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());
		
			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());
			
			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}
		
		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, 
		class _12, class _13, class _14, class _15, class _16, class _17>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, 
				_12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, 
		class _12, class _13, class _14, class _15, class _16, class _17, class _18>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, 
				_12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}
		
		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, 
		class _12, class _13, class _14, class _15, class _16, class _17, class _18, class _19, class _20, class _21>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, 
				_12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18, _19 arg19, _20 arg20, _21 arg21, 
				work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20, arg21, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, 
		class _12, class _13, class _14, class _15, class _16, class _17, class _18, class _19, class _20, class _21, class _22, class _23>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, 
				_12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18, _19 arg19, _20 arg20, _21 arg21, 
				_22 arg22, work_unit* parent)
		{
			int q;
			if (parent != NULL) {
				q = parent->get_q();
			}
			else {
				q = q_assign++ % num_threads;
			}

			slab_object task = work_cache.template allocate<T>(q);
			T* wrk = static_cast<T*>(task.get_object());

			wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20, arg21, arg22, static_cast<T*>(parent));
			internal_spawn(task, q);

			return task;
		}

		template <class T>
		void spawn_for(int start, int stop, loop_policy lp)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init();
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17, arg18);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17, arg18,
						arg19);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17, arg18,
						arg19, arg20);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20, class _21>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20, _21 arg21)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17, arg18,
						arg19, arg20, arg21);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20, class _21, class _22>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20, _21 arg21, _22 arg22)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17, arg18,
						arg19, arg20, arg21, arg22);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		template <class T, class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20, class _21, class _22, class _23>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20, _21 arg21, _22 arg22, _23 arg23)
		{
			int chunk = (stop - start) / num_threads;
			int chunk_rem = (stop - start) % num_threads; 

			int i = start;
			for (int q = 0; q < num_threads; ++q) {
				slab_object task = work_cache.template allocate<T>(q);
				T* wrk = static_cast<T*>(task.get_object());
			
				wrk->work_init(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9,
						arg10, arg11, arg12, arg13, arg14, arg15, arg16, arg17, arg18,
						arg19, arg20, arg21, arg22, arg23);
				if (q == num_threads - 1) {
					wrk->bounds_init(i, i + chunk + chunk_rem);
				}
				else {
					wrk->bounds_init(i, i + chunk);
				}
				internal_spawn(task, q);
				i += chunk;
			}
		}

		slab_object spawn()
		{
			return spawn<Work>();
		}

		template <class _1>
		slab_object spawn(_1 arg1)
		{
			return spawn<Work>(arg1);
		}

		template <class _1, class _2>
		slab_object spawn(_1 arg1, _2 arg2)
		{
			return spawn<Work>(arg1, arg2);
		}

		template <class _1, class _2, class _3>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3)
		{
			return spawn<Work>(arg1, arg2, arg3);
		}

		template <class _1, class _2, class _3, class _4>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4);
		}

		template <class _1, class _2, class _3, class _4, class _5>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7)
		{
			spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
		}
	
		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, class _12>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, _12 arg12)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, class _12,
			class _13, class _14, class _15, class _16, class _17, class _18>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, _12 arg12,
				_13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, 
					arg13, arg14, arg15, arg16, arg17, arg18);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, class _12,
			class _13, class _14, class _15, class _16, class _17, class _18, class _19>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, _12 arg12,
				_13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18, _19 arg19)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, 
					arg13, arg14, arg15, arg16, arg17, arg18, arg19);
		}
		
		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, class _12,
			class _13, class _14, class _15, class _16, class _17, class _18, class _19, class _20, class _21, class _22, class _23>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, _12 arg12,
				_13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18, _19 arg19, _21 arg21, _22 arg22)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, 
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20, arg21, arg22);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9, class _10, class _11, class _12,
			class _13, class _14, class _15, class _16, class _17, class _18, class _19, class _20, class _21, class _22, class _23>
		slab_object spawn(_1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9, _10 arg10, _11 arg11, _12 arg12,
				_13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18, _19 arg19, _21 arg21, _22 arg22, _23 arg23)
		{
			return spawn<Work>(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, 
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20, arg21, arg22, arg23);
		}

		void spawn_for(int start, int stop, loop_policy lp)
		{
			return spawn_for<Work>(start, stop, lp);
		}

		template <class _1>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1)
		{
			return spawn_for<Work>(start, stop, lp, arg1);
		}

		template <class _1, class _2>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2);
		}

		template <class _1, class _2, class _3>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3);
		}

		template <class _1, class _2, class _3, class _4>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4);
		}

		template <class _1, class _2, class _3, class _4, class _5>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, arg19);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20, class _21>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20, _21 arg21)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20, arg21);
		}
		
		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20, class _21, class _22>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20, _21 arg21, _22 arg22)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20, arg21, arg22);
		}

		template <class _1, class _2, class _3, class _4, class _5, class _6, class _7, class _8, class _9,
			class _10, class _11, class _12, class _13, class _14, class _15, class _16, class _17, class _18,
			class _19, class _20, class _21, class _22, class _23>
		void spawn_for(int start, int stop, loop_policy lp, _1 arg1, _2 arg2, _3 arg3, _4 arg4, _5 arg5, _6 arg6, _7 arg7, _8 arg8, _9 arg9,
				_10 arg10, _11 arg11, _12 arg12, _13 arg13, _14 arg14, _15 arg15, _16 arg16, _17 arg17, _18 arg18,
				_19 arg19, _20 arg20, _21 arg21, _22 arg22, _23 arg23)
		{
			return spawn_for<Work>(start, stop, lp, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12,
					arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20, arg21, arg22, arg23);
		}

};

template <class Work>
factory<Work>::factory(const unsigned int request_q, const policy p)
{
	construct(request_q, get_num_cpus() / request_q, p);
}

template <class Work>
factory<Work>::factory(const unsigned int request_q, const unsigned int request_threads, const policy p)
{
	construct(request_q, request_threads, p);
}

template <class Work>
void factory<Work>::construct(const unsigned int request_q, const unsigned int request_threads, const policy p)
{
	num_cpus 		= get_num_cpus();
	q_type 			= request_q;
	num_threads		= request_threads;
	work 			= new WorkQueue[num_threads];
	work_access	 	= new FLock[num_threads];
	q_assign 		= 0;

	gone_to_sleep		= new volatile bool[num_threads];
	done			= new bool[num_threads];
	not_joined 		= true;
	not_sleeped 		= true;

	ref_count		= 0;
	sense			= false;
	local_sense		= new volatile bool[num_threads];

	//added = executed = 0;

	work_cache.init(num_threads);

	switch (p) {
		case LOCAL:	policy_alg = local; break;
		case STEAL:	policy_alg = steal; break;
		case STEAL_SMT:	policy_alg = steal_smt; break;
	}

	for (unsigned int i = 0; i < num_threads - 1; ++i) {
		threads.push_back(new Thread());
	}
}

template <class Work>
factory<Work>::~factory()
{
	for (typename ThreadList::iterator i = threads.begin(); i != threads.end(); ++i) {
		delete *i;
	}
	
	delete [] work;
	delete [] work_access;
	delete [] gone_to_sleep;
	delete [] done;

	//std::cout << "added: " << added << " executed: " << executed << std::endl << std::flush;
}

template <class Work>
inline void factory<Work>::internal_spawn(slab_object task, int q)
{
	//atmc_add(&added, 1);

	task->set_q(q);
	work_access[q].lock();
		work[q].push(task);
	work_access[q].unlock();
}

template <class Work>
void factory<Work>::start_working()
{
	not_joined = true;
	wakeup();

	unsigned int cpu_count = 2;
	unsigned int q_count = 1;
	int delta = 2;
	for (typename ThreadList::iterator i = threads.begin(); i != threads.end(); ++i) {

		(*i)->create(factory_adapter(this), cpu_count, q_count);
		++q_count;

		cpu_count = (cpu_count + delta) % num_cpus;
		if (cpu_count == 0) {
			cpu_count = num_cpus - 1;
			delta = -delta;
		}
	}

	work_cache.local_init();
}

template <class Work>
inline void factory<Work>::sleep()
{
	bool all_asleep = false;

	not_sleeped = false;

	all_asleep = gone_to_sleep[0];
	for (int q = 1; q < num_threads; ++q) {
		all_asleep = all_asleep && gone_to_sleep[q];
	}

	if (!all_asleep) {
		long_sleep();
	}
}

template <class Work>
void factory<Work>::long_sleep()
{
	bool all_asleep = false;

	while (!all_asleep) {

		sched_yield();

		all_asleep = gone_to_sleep[0];
		for (int q = 1; q < num_threads; ++q) {
			all_asleep = all_asleep && gone_to_sleep[q];
		}
	}
}

template <class Work>
inline void factory<Work>::wakeup()
{
	not_sleeped = true;
}

template <class Work>
inline void factory<Work>::stop_working()
{
	not_joined = false;
	std::for_each(threads.begin(), threads.end(), std::mem_fun(&Thread::join));
}

template <class Work>
inline void factory<Work>::change_policy(const policy p)
{
	switch (p) {
		case LOCAL:	policy_alg = local; break;
		case STEAL:	policy_alg = steal; break;
		case STEAL_SMT:	policy_alg = steal_smt; break;
	}
}

template <class Work>
inline int factory<Work>::get_num_threads()
{
	return num_threads;
}

template <class Work>
inline void factory<Work>::root_barrier(slab_object b)
{
	barrier(b, PARENT, 0);
}

template <class Work>
inline void factory<Work>::parent_barrier(slab_object b, int q)
{
	barrier(b, PARENT, q);
}

template <class Work>
inline void factory<Work>::work_end_barrier(tree_unit* t)
{
	slab_object task(t);
	barrier(task, CHILD, task->get_q());
}

template <class Work>
inline void factory<Work>::child_barrier(tree_unit* t)
{
	slab_object task(t);
	barrier(task, CHILD, task->get_q());
}

template <class Work>
inline void factory<Work>::child_barrier(slab_object b, int q)
{
	barrier(b, CHILD, q);
}

template <class Work>
void factory<Work>::barrier(slab_object b, const int condition, int q)
{
	while (b->outstanding_work() != condition) {

		slab_object task = policy_alg(work, work_access, q, num_threads); 

		// I think that "keep going" is the correct behavior for when the barrier
		// finds there is no work. It is possible for the work to have been created
		// but not yet get onto one of the queues. However, this means that if 
		// users code their barriers wrong, they may hang.
		if (task == NULL) {
			continue;
		}

		//atmc_add(&executed, 1);
		task->do_work();
		task.free();

		if (task == b) {
			break;
		}
	}
}

template <class Work>
inline void factory<Work>::simple_barrier(work_unit* t)
{
	local_sense[t->get_q()] = !local_sense[t->get_q()];

	if (atmc_fetch_and_add(&ref_count, 1) == num_threads - 1) {
		ref_count = 0;
		sense = local_sense[t->get_q()];
	}
	else {
		while (sense != local_sense[t->get_q()]) { }
	}
}

template <class Work>
void factory<Work>::barrier()
{
retry:	schedule(0, 0);

	for (unsigned int q = 0; q < num_threads; ++q) {
		//if (!work[q].empty()) {
		if (!done[q]) {
			goto retry;
		}
	}
}

template <class Work>
void factory<Work>::long_barrier()
{
	bool all_done = false;
	unsigned long count = 0;

	while (!all_done) {

		sched_yield();

		all_done = work[0].empty();
		for (unsigned int q = 1; q < num_threads; ++q) {
			all_done = all_done && work[q].empty();
		}
	}
}

template <class WorkQueue>
slab_object local(WorkQueue* work, FLock* work_access, unsigned int q, unsigned int num_qs) 
{
	slab_object task;

	work_access[q].lock();
		if (!work[q].empty()) {
			task = work[q].front();
			work[q].pop();
		}
	work_access[q].unlock();

	return task;
}

template <class WorkQueue>
slab_object steal(WorkQueue* work, FLock* work_access, unsigned int q, unsigned int num_qs)
{
	slab_object task;
	bool steal_work = false;

	work_access[q].lock();
		if (!work[q].empty()) {
			task = work[q].front();
			work[q].pop();

			if (task == NULL) {
				steal_work = true;
			}
		}
		else {
			steal_work = true;
		}
	work_access[q].unlock();

	if (steal_work) {

		bool found_work = false;
		for (unsigned int other = (q + 1) % num_qs; other != q; other = (other + 1) % num_qs) {

			work_access[other].lock();
				if (!work[other].empty()) {
					task = work[other].front();
					work[other].pop();

					if (task != NULL) {
						task->set_q(q);
						found_work = true;
					}
				}
			work_access[other].unlock();

			if (found_work) {
				break;
			}
		}
	}

	return task;
}

template <class WorkQueue>
slab_object steal_smt(WorkQueue* work, FLock* work_access, unsigned int q, unsigned int num_qs)
{
	slab_object task;

	bool steal_work = false;

	work_access[q].lock();
		if (!work[q].empty()) {
			task = work[q].front();
			work[q].pop();
		}
		else {
			steal_work = true;
		}
	work_access[q].unlock();

	if (steal_work) {

		bool found_work = false;
		unsigned int neighbor;
		unsigned int stop;
		if (q % 2 == 0) {
			neighbor = (q - 1) % num_qs;
			stop = (q - 2) % num_qs;
		}
		else {
			neighbor = (q + 1) % num_qs;
			stop = q;
		}
		int count = 0;
		for (unsigned int other = neighbor; other != stop; other = (other + 1) % num_qs) {

			if (++count % 1000000000 == 0) {
				std::cout << other << ":" << neighbor << ":" << stop << std::endl << std::flush;
			}

			if (other == q) {
				continue;
			}
			work_access[other].lock();
				if (!work[other].empty()) {
					task = work[other].front();
					work[other].pop();

					if (task != NULL) {
						task->set_q(q);
						found_work = true;
					}
				}
			work_access[other].unlock();

			if (found_work) {
				break;
			}
		}
	}

	return task;
}

template <class Work>
void factory<Work>::schedule(const unsigned int cpu, const unsigned int q)
{
	//bool done = false;
	slab_object task;

	bindme_physical_cpu(cpu);
	//work_cache.local_init();

	gone_to_sleep[q] = false;
	done[q] = false;
	while (not_joined || !done[q]) {

		if (!not_sleeped) {
			gone_to_sleep[q] = true;
			continue;
		}
		gone_to_sleep[q] = false;

		task = policy_alg(work, work_access, q, num_threads);

		if (task == NULL) {
			done[q] = true;

			if (q == 0) {
				for (unsigned int i = 0; i < num_threads; ++i) {
					//if (!work[q].empty()) {
					if (!done[i]) {
						continue;
					}
				}
				break;
			}
			continue;
		}
		done[q] = false;

		bool freeable = !task->is_root();

		//atmc_add(&executed, 1);
		task->do_work();
		if (freeable) {
			task.free();
		}
	}
	gone_to_sleep[q] = true;
}

#endif	// __FACTORY_H__

