/*
 * slab_manager.h
 *
 * Scott Schneider
 */

#ifndef __SLAB_ALLOCATOR_H__
#define __SLAB_ALLOCATOR_H__

#include <vector>
#include <cassert>
#include <iostream>
#include <memory>
#include <cstdlib>
#include "locks.h"
#include "work_unit.h"
#include "mem_manager.h"

class lf_allocator
{
	typedef std::vector<pool_handler_t*> ManagerList;

	private:
		ManagerList	managers;
		unsigned int	num_cpus;

	public:
		lf_allocator(): num_cpus(0) {}
		~lf_allocator();

		void init(const unsigned int num_cpus);
		void local_init();

		template <class T>
		void register_class();

		template <class T>
		slab_object allocate(const unsigned int q);
};

class slab_object
{	
	friend class lf_allocator;

	private:
		work_unit*	object;

	public:
		slab_object(): object(NULL) {}
		slab_object(const slab_object& source): object(source.object) {}
		slab_object(work_unit* p): object(p) { }

		void operator=(work_unit* ptr)
		{
			object = ptr;
		}

		void operator=(const slab_object& source)
		{
			object = source.object;
		}

		int get_index()			{ return object->slab_index; }
		work_unit* get_object()		{ return object; }
		slab* get_slab()		{ return object->slab_ptr; }

		void set_index(int i)		{ object->slab_index = i; }
		void set_slab(slab* s)		{ object->slab_ptr = s; }

		bool operator==(const slab_object& s)	{ return object == s.object; }
		bool operator!=(const slab_object& s)	{ return object != s.object; }
		bool operator==(work_unit* ptr)		{ return object == ptr; }
		bool operator!=(work_unit* ptr)		{ return object != ptr; }
		work_unit* operator->() 		{ return object; }

		void free()
		{
			if (object->slab_ptr) {
				pool_deallocate(reinterpret_cast<pool_handler_t*>(object->slab_ptr), object);
			}
			object->slab_ptr = NULL;
		}
};

inline lf_allocator::~lf_allocator()
{
	for (ManagerList::iterator i = managers.begin(); i != managers.end(); ++i) {
		/*
		 delete_pool(*i);
		 */
	}
}

inline void lf_allocator::init(const unsigned int cpus)
{
	num_cpus = cpus;
}

inline void lf_allocator::local_init()
{
	mem_manager_thread_declare();
}

template <class T>
inline void lf_allocator::register_class()
{
	pool_handler_t* handler;
	init_pool(&handler, sizeof(T));
	managers.push_back(handler);
}

template <class T>
inline slab_object lf_allocator::allocate(const unsigned int q)
{
	slab_object so;
	pool_handler_t* handler = NULL;
	unsigned int object_size = sizeof(T);

	for (ManagerList::const_iterator i = managers.begin(); i != managers.end(); ++i) {
		if (object_size == (*i)->true_object_size) {
			handler = *i;
		}
	}

	assert(handler != NULL);

	so.object = reinterpret_cast<work_unit*>(pool_allocate(handler));
	assert(so.object != NULL);

	new (so.object) T();
	assert(so.object != NULL);

	so->slab_ptr = reinterpret_cast<slab*>(handler);
	return so;
}

#endif	// __SLAB_ALLOCATOR_H__

