#ifndef __MEM_MANAGER_H__
#define __MEM_MANAGER_H__

#include <cstdlib>
#include <unistd.h>

#include "c_queues.h"
#include "atomic.h"

#include <sys/types.h>
#include <linux/unistd.h>
              
_syscall0(pid_t,gettid)
pid_t gettid(void);

#define PAGESIZE	4096
#define SLAB_SIZE	4*PAGESIZE
#define CACHE_LINE_SIZE	64
#define MAX_THREADS	64

struct slab_type {
	char pool[SLAB_SIZE - sizeof(struct slab_type *)];
	struct slab_type *next;
};

typedef struct slab_type slab_t;

/* __thread_num is a thread-local variable */
static __thread int __thread_num;

/* __counter is the number of threads using the memory manager */
static volatile unsigned long __counter = 0;

typedef struct {
	/* The union ensures proper alignement for the cpmxchg_8b operations */
 	union {
		/* slab_limit[0] is the first free slot *
 		 * slab_limit[1] is the end of the slab */
		volatile unsigned long limit[3];
		volatile unsigned long long limits;
	} slab;
	volatile slab_t *top_slab;
	unsigned long long _pad1[8];
	lf_lifo_queue_t *free_queue[MAX_THREADS][CACHE_LINE_SIZE/sizeof(lf_lifo_queue_t *)];
	unsigned long long _pad2[8];
	int object_size;
	int true_object_size;
} pool_handler_t;

static __inline__ void mem_manager_thread_declare(void);
static __inline__ void init_pool(pool_handler_t **handler, int object_size);
static __inline__ void delete_pool(pool_handler_t *handler);
static __inline__ void *pool_allocate(pool_handler_t *handler);
static __inline__ void pool_deallocate(pool_handler_t *handler, void *object);

/******************************************************************************/

static __inline__ void mem_manager_thread_declare(void)
{
	__thread_num = atmc_fetch_and_add(&__counter, 1);
	fprintf(stdout, "tid = %d,  __thread_num = %d, &__thread_num = %p\n",  gettid(), __thread_num, &__thread_num);
	fflush(stdout);
}

static __inline__ void init_pool(pool_handler_t **handler, int object_size)
{
	int i;
	
	/* Allocate the memory pool handler */ 
	*handler = (pool_handler_t *)malloc(sizeof(pool_handler_t));
	if (handler == NULL) {
		fprintf(stderr, "Error: Could not allocate memory pool handler\n");
		fflush(stderr);
		exit(1);
	}

	/* Calculate the object size as a multiple of cache line */
	(*handler)->true_object_size = object_size;
	(*handler)->object_size = ceil((float)(object_size /*+ sizeof(pool_handler_t*)*/)/(float)CACHE_LINE_SIZE) * CACHE_LINE_SIZE;
	/* Initialize the queues */
	for (i = 0; i < MAX_THREADS; i++)
		lf_lifo_queue_init(&((*handler)->free_queue[i][0]));
	/* Allocate and initialize the 1st slab */
	if (posix_memalign((void **)(&((*handler)->top_slab)), PAGESIZE, SLAB_SIZE) != 0) {
		fprintf(stderr, "Error: Could not allocate a new slab\n");
		fflush(stderr);
		exit(1);
	}
	fprintf(stdout, "psm: %p\n", (*handler)->top_slab);
	fflush(stdout);
	(*handler)->top_slab->next = NULL;
	(*handler)->slab.limit[2] = (unsigned long)&((*handler)->top_slab->pool);
	(*handler)->slab.limit[0] = (unsigned long)&((*handler)->top_slab->pool);
	(*handler)->slab.limit[1] = (unsigned long)&((*handler)->top_slab->pool) + 
		(((unsigned long)&((*handler)->top_slab->next) - 
		  (unsigned long)&((*handler)->top_slab->pool)) / (*handler)->object_size) *
		(*handler)->object_size;
}

/******************************************************************************/

static __inline__ void delete_pool(pool_handler_t *handler)
{
	slab_t *slab, *next_slab;
	int i;
	
	for (i = 0; i < MAX_THREADS; i++)
		queue_end(handler->free_queue[__thread_num][0]);
	slab = (slab_t *)handler->top_slab;
	while(slab) {
		next_slab = slab->next;
		free(slab);
		slab = next_slab;
	}
	free(handler);
}

/******************************************************************************/

static void *pool_allocate(pool_handler_t *handler)
{
	unsigned long pointer, next_index, slab_end;
	slab_t *slab;
	int target_thread;

	/*
	fprintf(stderr, "Thread %d allocating\n", __thread_num);
	fflush(stderr);
	*/

//	fprintf(stdout, "%d:%d:%p\n", gettid(), *(int *)((unsigned long)(&__thread_num) - 4), &__thread_num);
	fprintf(stdout, "%d:%d:%p\n", gettid(), __thread_num, &__thread_num);
	fflush(stdout);
	//assert(__thread_num == 0);
	/* Try to allocate from the freed FIFO first */
	pointer = (unsigned long)lf_lifo_dequeue(handler->free_queue[__thread_num][0]);
	if ((void *)pointer != NULL) {

		/*
		printf("Allocated %p\n", (void *)pointer);
		*/

		assert(pointer != NULL);
		/*
		*((pool_handler_t**)pointer) = handler;
		pointer += sizeof(pool_handler_t*);
		*/
		return ((void *)pointer);
	}
	/* Try to allocate from the slab */
	do {
		pointer = handler->slab.limit[0];
		slab_end = handler->slab.limit[1];

		/*
		printf("pointer %p slab_end %p\n", pointer, slab_end);
		*/

		/* The slab is over, however someone else will allocate a new one */
		if (pointer > slab_end)
			continue;
		/* The slab is over and I will be probably the one to expand it. Let's 	*
		 * first try to find memory in the freed queues of other threads	*/
		else if (pointer == slab_end) {

			/*
			printf("Thread %d searching other queues\n", __thread_num);
			*/

			target_thread = __thread_num + 1;
			while(target_thread % (__counter+1) != __thread_num) {
				next_index = (unsigned long)lf_lifo_dequeue(handler->free_queue[target_thread][0]);
				if ((void *)next_index != NULL) {

					/*
					printf("Allocated %p\n", (void *)next_index);
					*/

					assert(next_index != NULL);
					/*
					if (next_index >= handler->slab.limit[2] && next_index <= handler->slab.limit[1]) {
						fprintf(stderr, "%p:%p:%p:%d:\n", next_index, handler->slab.limit[2], handler->slab.limit[1], __thread_num));
						fflush(stderr);
					}
					*/
					/*
					*((pool_handler_t**)next_index) = handler;
					return ((void *)(next_index + sizeof(pool_handler_t*)));
					*/
					return ((void *)(next_index));
				}
				target_thread++;
			}
			/*
			printf("Thread %d found other queues empty\n", __thread_num);
			*/
		}
		next_index = pointer + handler->object_size;
		/*
		printf("next_index = %p\n", next_index);
		*/
	}
	while (!compare_and_swap8b(&(handler->slab.limits), pointer, slab_end, 
				   next_index, slab_end));
	if (pointer != slab_end) {
		/*
		printf("Thread %d returning memory from the slab\n", __thread_num);
		printf("Allocated %p\n", (void *)pointer);
		*/
		assert(pointer != NULL);
		/*
		if (next_index >= handler->slab.limit[2] && next_index <= handler->slab.limit[1]) {
			fprintf(stderr, "%p:%p:%p:%d:\n", next_index, handler->slab.limit[2], handler->slab.limit[1], __thread_num);
			fflush(stderr);
		}
		*/
		/*
		*((pool_handler_t**)pointer) = handler;
		pointer += sizeof(pool_handler_t*);
		*/
		return((void *)pointer);
	}

	/*
	printf("Thread %d allocating new slab\n", __thread_num);
	*/

	/* I am the one to allocate the new slab!!! */
	if (posix_memalign((void **)(&slab), PAGESIZE, SLAB_SIZE) != 0) {
		fprintf(stderr, "Error: Could not allocate a new slab\n");
		fflush(stderr);
		exit(1);
	}

	slab->next = (slab_t *)handler->top_slab;
	handler->top_slab = slab;
	pointer = (unsigned long)(handler->top_slab->pool);

	/* Keep the 1st slot of the slab for the object allocation that drove me *
	 * into this function... 						 */
	next_index = pointer + handler->object_size; 
	slab_end = (unsigned long)(handler->top_slab->pool) + 
		   (((unsigned long)&(handler->top_slab->next) - 
		   (unsigned long)(handler->top_slab->pool)) / handler->object_size) *
		   handler->object_size;
	handler->slab.limit[2] = pointer;
	/* Change the slab limits to the new ones */
	while(!compare_and_swap8b(&(handler->slab.limits), handler->slab.limit[0], 
				  handler->slab.limit[1], next_index, slab_end));
	/*
	fprintf(stderr, "%p:%p:%d:\n", next_index, slab_end, __thread_num);
	fflush(stderr);
	*/

	/*
	printf("Allocated %p\n", (void *)pointer);
	*/
	assert(pointer != NULL);
	/*
	if (next_index >= handler->slab.limit[2] && next_index <= handler->slab.limit[1]) {
		fprintf(stderr, "%p:%p:%p:%d:\n", next_index, handler->slab.limit[2], handler->slab.limit[1], __thread_num);
		fflush(stderr);
	}
	*/
	/*
	*((pool_handler_t**)pointer) = handler;
	pointer += sizeof(pool_handler_t*);
	*/
	return((void *)pointer);	
}

/******************************************************************************/

static __inline__ void pool_deallocate(pool_handler_t *handler, void *object)
{
	/*
	printf("Thread %d deallocating object %p\n", __thread_num, object);
	*/
	
	if (object != NULL) {
		/*
		object -= sizeof(pool_handler_t*);
		assert(*(pool_handler_t**)object == handler);
		lf_lifo_enqueue((*((pool_handler_t**)object))->free_queue[__thread_num][0], object);
		*/
		lf_lifo_enqueue(handler->free_queue[__thread_num][0], object);
	}
}

#endif /*__MEM_MANAGER_H__*/

