/*
 * posix_thread.h
 * 
 * Scott Schneider
 */

#ifndef __POSIX_THREAD_H__
#define __POSIX_THREAD_H__

#include <iostream>
#include <ostream>
#include <functional>
#include <cassert>
#include <pthread.h>

class mutex
{
	friend class condition;

	private:
		pthread_mutex_t m;

	public:
		mutex() { pthread_mutex_init(&m, NULL);	}
		~mutex() { pthread_mutex_destroy(&m); }
		
		void lock() { pthread_mutex_lock(&m); }
		void unlock() {	pthread_mutex_unlock(&m); }
};

class condition
{
	friend class mutex;

	private:
		pthread_cond_t c;

	public:
		condition() { pthread_cond_init(&c, NULL); }
		~condition() { pthread_cond_destroy(&c); }
		
		void wait(mutex& lock) { pthread_cond_wait(&c, &lock.m); }
		void signal() { pthread_cond_signal(&c); }
};

template <class Arguments>
class posix_thread;

template <class Arguments>
class attribute
{
	friend class posix_thread<Arguments>;

	private:
		pthread_attr_t attr;

	public:
		attribute() { pthread_attr_init(&attr);	}
		
		void schedule_fifo() { pthread_attr_setschedpolicy(&attr, SCHED_FIFO); }
		void schedule_rr() { pthread_attr_setschedpolicy(&attr, SCHED_RR); }
		void schedule_other() { pthread_attr_setschedpolicy(&attr, SCHED_OTHER); }
		void schedule_inherit() {pthread_attr_setinheritsched(&attr, PTHREAD_INHERIT_SCHED); }
		void schedule_explicit() { pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED); }
		
		void set_priority(int priority)
		{
			sched_param pr;
			pr.sched_priority = priority;

			pthread_attr_setschedparam(&attr, &pr);		
		}
};

template <class Arguments>
class posix_thread
{
	private:
		pthread_t	th;
		pthread_attr_t*	attr;
		bool		running;

		template <class Function>
		void call_pthread_create(Function f, Arguments* args)
		{
			// If a create fails, what should we do? For now, we'll just make sure
			// the the thread knows it's not running; passing this information on 
			// up to the caller (i.e., throw an exception) doesn't seem like 
			// a good course of action - I don't think a failed create should be 
			// a show stopper. Or should it?
			if (int ret = pthread_create(&th, attr, (void*(*)(void*))f, args)) {
				running = false;
		
				std::cerr << "posix_thread::call_pthread_create failure: ";
				switch (ret) {
					case EAGAIN: std::cerr << "EAGAIN";
						break;
					case EINVAL: std::cerr << "EINVAL";
						break;
					case EPERM:  std::cerr << "EPERM";
						break;
					case ENOMEM: std::cerr << "ENOMEM";
						break;
					default:     std::cerr << "Error unknown";
				}
				std::cerr << std::endl << std::flush;
			}
			else {
				running = true;
				assert(th != 0);
			}
		}

		static int wrapper_1(Arguments* args)
		{
			args->function(args->_1);
			delete args;
			return 0;
		}

		static int wrapper_2(Arguments* args)
		{
			args->function(args->_1, args->_2);
			delete args;
			return 0;
		}

		static int wrapper_3(Arguments* args)
		{
			args->function(args->_1, args->_2, args->_3);
			delete args;
			return 0;
		}

		static int wrapper_4(Arguments* args)
		{
			args->function(args->_1, args->_2, args->_3, args->_4);
			delete args;
			return 0;
		}

		static int wrapper_5(Arguments* args)
		{
			args->function(args->_1, args->_2, args->_3, args->_4, args->_5);
			delete args;
			return 0;
		}

		static int wrapper_6(Arguments* args)
		{
			args->function(args->_1, args->_2, args->_3, args->_4, args->_5, args->_6);
			delete args;
			return 0;
		}

		static int wrapper_7(Arguments* args)
		{
			args->function(args->_1, args->_2, args->_3, args->_4, args->_5, args->_6, args->_7);
			delete args;
			return 0;
		}

	public:
		posix_thread()
		{
			attr = NULL;
			running = false;
		}
		
		posix_thread(attribute<Arguments> atrobj)
		{
			attr = new pthread_attr_t(atrobj.attr);
			running = false;
		}
			
		posix_thread(const posix_thread<Arguments>& source)
		{
			th = source.th;
			running = source.running;
		
			if (source.attr == NULL) {
				attr = NULL;
			}
			else {
				attr = new pthread_attr_t(*source.attr);
			}		
		}
		
		posix_thread& operator=(const posix_thread<Arguments>& source)
		{
			th = source.th;
			running = source.running;
		
			if (source.attr == NULL) {
				attr = NULL;
			}
			else {
				attr = new pthread_attr_t(*source.attr);
			}
		
			return *this;		
		}

		template <class Return>
		void create(Return (*f)())
		{
			call_pthread_create<Return (*)(), void>(f, NULL);		
		}
		
		template <class Return>
		void create(Return (*f)(Arguments*), Arguments* arg1)
		{
			call_pthread_create(f, arg1);		
		}

		template <class Function,
			class _1>
		void create(Function f,
			_1 arg1)
		{
			Arguments temp = { f, arg1 };
			Arguments* args = new Arguments(temp);
			call_pthread_create(wrapper_1, args);		
		}

		template <class Function,
			class _1,
			class _2>
		void create(Function f, 
			_1 arg1, 
			_2 arg2)
		{
			Arguments temp = { f, arg1, arg2 };
			Arguments* args = new Arguments(temp);
			call_pthread_create(wrapper_2, args);		
		}

		template <class Function,
			class _1,
			class _2,
			class _3>
		void create(Function f, 
			_1 arg1, 
			_2 arg2, 
			_3 arg3)
		{
			Arguments temp = { f, arg1, arg2, arg3 };
			Arguments* args = new Arguments(temp);
			call_pthread_create(wrapper_3, args);		
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4>
		void create(Function f, 
			_1 arg1, 
			_2 arg2, 
			_3 arg3, 
			_4 arg4)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4 };
			Arguments* args = new Arguments(temp);
			call_pthread_create(wrapper_4, args);		
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4,
			class _5>
		void create(Function f, 
			_1 arg1, 
			_2 arg2, 
			_3 arg3, 
		 	_4 arg4, 
			_5 arg5)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4, arg5 };
			Arguments* args = new Arguments(temp);
			call_pthread_create(wrapper_5, args);		
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4,
			class _5,
			class _6>
		void create(Function f,
			_1 arg1, 
			_2 arg2, 
			_3 arg3, 
			_4 arg4, 
			_5 arg5, 
			_6 arg6)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4, arg5, arg6, arg7 };
			Arguments* args = new Arguments(temp);
			call_pthread_create(wrapper_6, args);		
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4,
			class _5,
			class _6,
			class _7>
		void create(Function f,
			_1 arg1,
			_2 arg2,
			_3 arg3,
			_4 arg4,
			_5 arg5,
			_6 arg6,
			_7 arg7)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4, arg5, arg6, arg7 };
			Arguments* args = new Arguments(temp);
			call_pthread_create(wrapper_7, args);		
		}
	
		void join()
		{
			void* thread_return;
		
			// we should probably do something with the return code...
			if (running) {
				assert(th != 0);
				if (int ret = pthread_join(th, &thread_return)) {

					std::cerr << "posix_thread::pthread_join failure: ";
					switch(ret) {
						case ESRCH:	std::cerr << "ESRCH.";
								break;
						case EINVAL:	std::cerr << "EINVAL.";
								break;
						case EDEADLK:	std::cerr << "EDEADLK.";
								break;
					}
					std::cerr << std::endl << std::flush;
				}
				running = false;
			}		
		}
		
		bool active() const { return running; }

		void schedule_fifo(int priority)
		{
			sched_param pr;
			pr.sched_priority = priority;
		
			pthread_setschedparam(th, SCHED_FIFO, &pr);		
		}
		
		void schedule_rr(int priority)
		{
			sched_param pr;
			pr.sched_priority = priority;
		
			pthread_setschedparam(th, SCHED_RR, &pr);		
		}
		
		void schedule_other(int priority)
		{
			sched_param pr;
			pr.sched_priority = priority;
		
			pthread_setschedparam(th, SCHED_OTHER, &pr);		
		}

		bool operator<(const posix_thread<Arguments>& other) { return th < other.th; }
		
		std::ostream& operator<<(std::ostream& out)
		{
			out << th;
			return out;		
		}
};

/*
 * Needed for a hash map. Taken from "The C++ Programming Language", Bjarne
 * Stroustrup, page 503.
 */
template <class T>
struct hash_fun {
	size_t operator()(T* ptr) const
	{
		size_t res = 0;

		size_t len = sizeof(T*);
		const char* p = reinterpret_cast<const char*>(ptr);

		while (len--) {
			res = (res << 1) ^* p++;
		}
		return res;
	}
};

#endif	// __POSIX_THREAD_H__
