/*
 * slab_manager.h
 *
 * Scott Schneider
 */

#ifndef __SLAB_ALLOCATOR_H__
#define __SLAB_ALLOCATOR_H__

#include <vector>
#include <list>
#include <cassert>
#include <iostream>
#include <memory>
#include <cstdlib>
#include "locks.h"
#include "work_unit.h"

const bool OBJECT_FREE  = false;
const bool OBJECT_TAKEN = true;

const unsigned int START_SLAB_SIZE = 1000;
const double GROWTH_RATE = 1.25;

typedef exp_spin_lock SLock;

class slab;
class slab_manager;
class general_manager;

enum slab_state { ALL_FREE, PARTIAL, ALL_TAKEN };

class slab
{
	private:
		char*		entries;
		bool*		entries_status;
		slab_state	status; 
		SLock		access;
		unsigned int	free_count;
		unsigned int	curr;
		unsigned int	slab_size;
		unsigned int	object_size;

	public:
		slab(int size, int object);
		~slab();

		template <class T>
		void init();

		slab_state availability();
		unsigned int get_object_size() const;
		unsigned int get_slab_size();
		slab_object allocate();
		void free(unsigned int);
};

class slab_manager
{	
	typedef std::list<slab*> SlabList;

	private:
		SlabList*	local_slabs;
		SLock*		slabs_access;
		unsigned int	num_cpus;
		unsigned int	next_slab_size;
		unsigned int	object_size;
		bool		initialized;

		template <class T>
		slab* get_free_slab(unsigned int q);

	public:
		slab_manager();
		~slab_manager();

		template <class T>
		void init(const unsigned int num_cpus);

		bool is_initialized();
		unsigned int get_object_size() const;

		template <class T>
		slab_object allocate(const unsigned int q);
};

class general_manager
{
	typedef std::vector<slab_manager*> ManagerList;

	private:
		ManagerList	managers;
		rw_lock		list_access;
		unsigned int	num_cpus;

	public:
		general_manager(): num_cpus(0) {}
		~general_manager();

		void init(const unsigned int num_cpus);
		void local_init() {}

		template <class T>
		void register_class();

		template <class T>
		slab_object allocate(const unsigned int q);
};

class slab_object
{	
	private:
		work_unit*	object;

	public:
		slab_object(): object(NULL) {}
		slab_object(const slab_object& source): object(source.object) {}
		slab_object(work_unit* p): object(p) { }

		void operator=(work_unit* ptr)			{ object = ptr; }
		void operator=(const slab_object& source)	{ object = source.object; }

		int get_index()			{ return object->slab_index; }
		work_unit* get_object()		{ return object; }
		slab* get_slab()		{ return object->slab_ptr; }

		void set_index(int i)		{ object->slab_index = i; }
		void set_slab(slab* s)		{ object->slab_ptr = s; }

		bool operator==(const slab_object& s)	{ return object == s.object; }
		bool operator!=(const slab_object& s)	{ return object != s.object; }
		bool operator==(work_unit* ptr)		{ return object == ptr; }
		bool operator!=(work_unit* ptr)		{ return object != ptr; }
		work_unit* operator->() 		{ return object; }

		void free()
		{
			/*
			if (object->slab_ptr) {
				object->slab_ptr->free(object->slab_index);
			}
			object->slab_ptr = NULL;
			object->slab_index = -1;
			*/
			delete object;
		}
};

inline slab::slab(int ss, int os):
	status(ALL_FREE), free_count(ss), curr(0), slab_size(ss), object_size(os)
{
	/*
	entries = new char[slab_size * object_size];
	entries_status = new bool[slab_size];

	memset(entries, 0, slab_size);
	memset(entries_status, OBJECT_FREE, slab_size);
	*/
}

template <class T>
inline void slab::init()
{
	entries = reinterpret_cast<char*>(new T[slab_size]);
	entries_status = new bool[slab_size];

	for (int i = 0; i < slab_size; ++i) {
		entries_status[i] = OBJECT_FREE;
	}
}
inline slab::~slab()
{
	//delete [] entries;
	delete [] entries_status;
}

inline slab_state slab::availability()
{
	return status;
}

inline unsigned int slab::get_object_size() const
{
	return object_size;
}

inline unsigned int slab::get_slab_size()
{
	return slab_size;
}

inline slab_object slab::allocate()
{
	slab_object so;

	access.lock();
		if (status != ALL_TAKEN) {

			// we have a guarantee that if there is availabile space in this
			// slab, then curr points to a free spot
			assert(entries_status[curr] == OBJECT_FREE);
			entries_status[curr] = OBJECT_TAKEN;
			so = reinterpret_cast<work_unit*>(&entries[curr * object_size]);
			so.set_slab(this);
			so.set_index(curr);
			--free_count;
			
			if (free_count > 0) {

				// maintain the guarantee that curr points to a free spot
				for (unsigned int i = (curr + 1) % slab_size; i != curr; i = (i + 1) % slab_size) {
					if (entries_status[i] == OBJECT_FREE) {
						curr = i;
						break;
					}
				}
				assert(entries_status[curr] == OBJECT_FREE);
				status = PARTIAL;
			}
			else {
				status = ALL_TAKEN;
				curr = slab_size;
			}
		}
		else {
			so = NULL;
		}
	access.unlock();

	return so;
}

inline void slab::free(const unsigned int i)
{
	access.lock();
		entries_status[i] = OBJECT_FREE;
		++free_count;

		if (free_count == slab_size) {
			status = ALL_FREE;
			curr = 0;
		}
		else {
			status = PARTIAL;
			if (i < curr) {
				curr = i;
			}
		}
		/*
		else if (free_count == 1) {
			curr = i;
			status = PARTIAL;
		}
		*/
	access.unlock();
}

inline slab_manager::slab_manager(): local_slabs(NULL), slabs_access(NULL), num_cpus(0), next_slab_size(0), object_size(0), initialized(false) {}

template <class T>
inline void slab_manager::init(const unsigned int cpus)
{
	num_cpus = cpus;
	object_size = sizeof(T);
	next_slab_size = START_SLAB_SIZE;

	local_slabs = new SlabList[num_cpus];
	slabs_access = new SLock[num_cpus];

	for (unsigned int i = 0; i < num_cpus; ++i) {
		slabs_access[i].lock();
			slab* s = new slab(next_slab_size, object_size);
			s->init<T>();
			local_slabs[i].push_front(s);
		slabs_access[i].unlock();
	}

	initialized = true;
}

inline slab_manager::~slab_manager()
{
	for (unsigned int i = 0; i < num_cpus; ++i) {
		for (SlabList::iterator it = local_slabs[i].begin(); it != local_slabs[i].end(); ++it) {
			delete *it;
		}
	}

	delete [] slabs_access;
	delete [] local_slabs;
}

inline bool slab_manager::is_initialized()
{
	return initialized;
}

inline unsigned int slab_manager::get_object_size() const
{
	return object_size;
}

template <class T>
inline slab* slab_manager::get_free_slab(unsigned int q)
{
	slab* slab_ptr = NULL;
	slabs_access[q].lock();

		/* As we search through the list, we move slabs to 
		 * the back if they're ALL_TAKEN. This helps to
		 * maintain a list order of {ALL_FREE, PARTIAL, ALL_TAKEN} so that
		 * the majority of the time, an ALL_FREE or PARTIAL slab is in
		 * the front.
		 *
		 * We also assume that the list has at least one slab. That's
		 * taken care of in the constructor, so that should be safe.
		 */
		SlabList::iterator i = local_slabs[q].begin();
		slab* first = local_slabs[q].front();
		do {
			if ((*i)->availability() != ALL_TAKEN) {
				slab_ptr = *i;
				break;
			}
			++i;
			local_slabs[q].splice(local_slabs[q].end(), local_slabs[q], local_slabs[q].begin());
		} while (i != local_slabs[q].end() && *i != first);

		if (slab_ptr == NULL) {
			slab_ptr = new slab(next_slab_size, object_size);
			slab_ptr->init<T>();
			local_slabs[q].push_front(slab_ptr);
			next_slab_size = (unsigned int)(next_slab_size * GROWTH_RATE);
		}
	slabs_access[q].unlock();

	return slab_ptr;
}

template <class T>
inline slab_object slab_manager::allocate(unsigned int q)
{
	slab_object so;

	while (so == NULL) {
		slab* s = get_free_slab<T>(q);
		so = s->allocate();
	}
	
	return so;
}

inline general_manager::~general_manager()
{
	for (ManagerList::iterator i = managers.begin(); i != managers.end(); ++i) {
		delete *i;
	}
}

inline void general_manager::init(const unsigned int cpus)
{
	num_cpus = cpus;
}

template <class T>
inline void general_manager::register_class()
{	
	slab_manager* sm = new slab_manager();
	sm->init<T>(num_cpus);
	managers.push_back(sm);
}

template <class T>
inline slab_object general_manager::allocate(const unsigned int q)
{
	slab_object so = new T();
	/*
	slab_manager* sm = NULL;
	unsigned int object_size = sizeof(T);

	for (ManagerList::const_iterator i = managers.begin(); i != managers.end(); ++i) {
		if (object_size == (*i)->get_object_size()) {
			sm = *i;
		}
	}

	so = sm->allocate<T>(q);
	*/
	return so;
}

#endif	// __SLAB_ALLOCATOR_H__

