/*
 * work_unit.h
 *
 * Scott Schneider
 */

#ifndef __WORK_UNIT_H__
#define __WORK_UNIT_H__

#include "atomic.h"

class slab;

class work_unit {

	friend class slab;
	friend class slab_object;
	friend class slab_manager;
	friend class general_manager;
	friend class lf_allocator;

	private:
		void*	next;		// needed by lockfree queues
		slab*	slab_ptr;
		int	slab_index;

	protected:
		int q;

	public:
		work_unit(): q(0), slab_ptr(NULL), slab_index(-1) { }
		virtual ~work_unit() {}

		void work_init()
		{
			q = -1;
		}

		void set_q(unsigned int the_q)
		{
			q = the_q;
		}

		int get_q()
		{
			return q;
		}

		virtual int outstanding_work() = 0;
		virtual void work() = 0;
		virtual void do_work() = 0;
		virtual bool is_root() = 0;
};

class tree_unit: public work_unit
{
	public:
		volatile unsigned long ref_count;
		tree_unit* parent;

		tree_unit(): ref_count(0), parent(NULL) {}

		virtual ~tree_unit() {}

		void work_init(tree_unit* p)
		{
			work_unit::work_init();

			ref_count = 1;
			parent = p;

			if (parent != NULL) {
				atmc_add(&parent->ref_count, 1);
			}
		}

		int outstanding_work()
		{
			return ref_count;
		}

		bool is_root()
		{
			return parent == NULL;
		}
};

class task_unit: public tree_unit
{
	public:
		void work_init(tree_unit* p)
		{
			tree_unit::work_init(p);
			if (p) {
				q = p->get_q();
			}
		}

		void do_work()
		{
			work();
			atmc_add(&ref_count, -1);

			if (parent != NULL) {
				atmc_add(&parent->ref_count, -1);
			}
		}
};

class loop_unit: public tree_unit
{
	protected:
		int loop_start;
		int loop_stop;

	public:
		void work_init(tree_unit* p)
		{
			tree_unit::work_init(p);
		}

		void bounds_init(int strt, int stp)
		{
			loop_start = strt;
			loop_stop = stp;
		}

		void do_work()
		{
			work();
			atmc_add(&ref_count, -1);

			if (parent != NULL) {
				atmc_add(&parent->ref_count, -1);
			}
		}
};

class plain_unit: public work_unit
{
	private:
		bool executed;
	public:
		plain_unit(): executed(false) {}

		virtual ~plain_unit() {}

		void work_init()
		{
			work_unit::work_init();
			executed = false;
		}

		int outstanding_work()
		{
			return (int)executed;
		}

		void do_work()
		{
			work();
			executed = true;
		}

		bool is_root()
		{
			return false;
		}
};

class mixed_work: public plain_unit
{
	int outstanding_work() { return -1;}
	void work() {}
};

#endif	// __WORK_UNIT_H__

